#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import TransformationSemantics
from ethtx.semantics.base import Base


class Weth(Base):
    code_hash = "0xd0a06b12ac47863b5c7be4185c2deaad1c61557033f56c7d4ea74429cbb25e23"
    contract_semantics = dict(
        transformations={
            "0xe1fffcc4923d04b559f4d29a8bfc6cda04eb5b0d3c460751c2402c5c5cc9109c": {  # Deposit
                "wad": TransformationSemantics(transformation="wad / 10**18")
            },
            "0x7fcf532c15f0a6db0bd6d0e038bea71d30d808c7d98cb3bf7268a95bf5081b65": {  # Withdrawal
                "wad": TransformationSemantics(transformation="wad / 10**18")
            },
            "0x2e1a7d4d": {  # withdraw
                "wad": TransformationSemantics(transformation="wad / 10**18")
            },
        }
    )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from copy import deepcopy
from typing import List, Dict, Optional, Union

import jsonpickle

from ethtx.utils.pickable import JsonObject


class TransformationSemantics:
    transformed_name: Optional[str]
    transformed_type: Optional[str]
    transformation: Optional[str]

    def __init__(
        self,
        transformed_name: Optional[str] = None,
        transformed_type: Optional[str] = None,
        transformation: Optional[str] = "",
    ):
        self.transformed_name = transformed_name
        self.transformed_type = transformed_type
        self.transformation = transformation


class ParameterSemantics:
    parameter_name: str
    parameter_type: str
    indexed: bool
    dynamic: bool
    components: list

    def __init__(
        self,
        parameter_name: str,
        parameter_type: str,
        components: list,
        indexed: bool = False,
        dynamic: bool = False,
    ):
        self.parameter_name = parameter_name
        self.parameter_type = parameter_type
        self.components = components
        self.indexed = indexed
        self.dynamic = dynamic


class EventSemantics:
    signature: str
    anonymous: bool
    name: str
    parameters: List[ParameterSemantics]

    def __init__(
        self,
        signature: str,
        anonymous: bool,
        name: str,
        parameters: List[ParameterSemantics],
    ):
        self.signature = signature
        self.anonymous = anonymous
        self.name = name
        self.parameters = parameters


class FunctionSemantics:
    signature: str
    name: str
    inputs: List[ParameterSemantics]
    outputs: List[ParameterSemantics]

    def __init__(
        self,
        signature: str,
        name: str,
        inputs: List[ParameterSemantics],
        outputs: List[ParameterSemantics],
    ):
        self.signature = signature
        self.name = name
        self.inputs = inputs
        self.outputs = outputs


class SignatureArg(JsonObject):
    name: str
    type: str

    def __init__(self, name: str, type: str):
        self.name = name
        self.type = type


class Signature(JsonObject):
    signature_hash: str
    name: str
    args: List[SignatureArg]
    count: int
    tuple: bool

    def __init__(
        self,
        signature_hash: str,
        name: str,
        args: List[SignatureArg],
        count: int = 1,
        tuple: bool = False,
    ):
        self.signature_hash = signature_hash
        self.name = name
        self.args = args
        self.count = count
        self.tuple = tuple


class ERC20Semantics:
    name: str
    symbol: str
    decimals: int

    def __init__(self, name: str, symbol: str, decimals: int):
        self.name = name
        self.symbol = symbol
        self.decimals = decimals


class ContractSemantics(JsonObject):
    code_hash: str
    name: str
    events: Dict[str, EventSemantics]
    functions: Dict[str, FunctionSemantics]
    transformations: [Dict[str, Dict[str, TransformationSemantics]]]

    def __init__(
        self,
        code_hash: str,
        name: str,
        events: Dict[str, EventSemantics],
        functions: Dict[str, FunctionSemantics],
        transformations: [Dict[str, Dict[str, TransformationSemantics]]],
    ):
        self.code_hash = code_hash
        self.name = name
        self.events = events
        self.functions = functions
        self.transformations = transformations


class AddressSemantics(JsonObject):
    chain_id: str
    address: str
    name: str
    is_contract: bool
    contract: Union[ContractSemantics, str]
    standard: Optional[str]
    erc20: Optional[ERC20Semantics]

    def __init__(
        self,
        chain_id: str,
        address: str,
        name: str,
        is_contract: bool,
        contract: ContractSemantics,
        standard: Optional[str],
        erc20: Optional[ERC20Semantics],
    ):
        self.chain_id = chain_id
        self.address = address
        self.name = name
        self.is_contract = is_contract
        self.contract = contract
        self.standard = standard
        self.erc20 = erc20

    def json_str(self, entire: Optional[bool] = True) -> str:
        """Return object as encoded json."""
        if entire:
            return jsonpickle.encode(self, unpicklable=False)

        new_obj = deepcopy(self)
        new_obj.contract = new_obj.contract.code_hash

        return jsonpickle.encode(new_obj, unpicklable=False)

    def json(self, entire: Optional[bool] = True) -> Dict:
        """Return object as decoded dict."""
        if entire:
            return jsonpickle.decode(self.json_str(entire))

        new_obj = deepcopy(self)
        new_obj.contract = new_obj.contract.code_hash
        return jsonpickle.decode(self.json_str(entire))



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import TransformationSemantics
from ethtx.semantics.base import Base


class CDPManager(Base):
    code_hash = "0xdc70464baedd675aadf1b109ac8ec36f78adb5db19bc087d2a70208305c786b1"
    contract_semantics = dict(
        transformations={
            "0x6090dec5": {  # open
                "ilk": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(ilk)"
                )
            },
            "0x45e6bdcd": {  # frob
                "dart": TransformationSemantics(transformation="dart / 10**18"),
                "dink": TransformationSemantics(transformation="dink / 10**18"),
            },
            "0xf9f30db6": {  # move
                "rad": TransformationSemantics(transformation="rad / 10**45")
            },
            "0x18af4d60": {  # flux
                "ilk": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(ilk)"
                ),
                "wad": TransformationSemantics(transformation="wad / 10**18"),
            },
            "0x2c2cb9fd": {  # ilks
                "__output0__": TransformationSemantics(
                    transformed_type="string",
                    transformation="string_from_bytes(__output0__)",
                )
            },
        }
    )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import TransformationSemantics
from ethtx.semantics.base import Base


class DsProxy(Base):
    code_hash = "0x27c02a1a822222c2ad6a9a01021c98abf05dbe6d19540035756ef97697ed41d0"
    contract_semantics = dict(
        transformations={
            "0x1cff79cd": {  # execute
                "_data": TransformationSemantics(
                    transformed_type="call",
                    transformation="decode_call(_target, _data)",
                )
            }
        }
    )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from abc import ABC, abstractmethod
from typing import Dict, List, Any, Iterator, TypedDict, Union, Tuple

import requests

from ethtx.exceptions import FourByteConnectionException

log = logging.getLogger(__name__)


class SignatureReturnType(TypedDict):
    name: str
    args: Union[List[str], Tuple[str]]


class SignatureProvider(ABC):
    @abstractmethod
    def list_function_signatures(self, filters: Dict):
        ...

    @abstractmethod
    def list_event_signatures(self, filters: Dict):
        ...

    @abstractmethod
    def get_function(self, signature: str):
        ...

    @abstractmethod
    def get_event(self, signature: str):
        ...


class FourByteProvider(SignatureProvider):
    API_URL: str = "https://www.4byte.directory/api/v1"
    FUNCTION_ENDPOINT: str = "signatures"
    EVENT_ENDPOINT: str = "event-signatures"

    def list_function_signatures(self, filters: Dict = None) -> List[Dict]:
        return self._get_all(endpoint=self.FUNCTION_ENDPOINT, filters=filters)

    def list_event_signatures(self, filters: Dict = None) -> List[Dict]:
        return self._get_all(endpoint=self.EVENT_ENDPOINT, filters=filters)

    def get_function(self, signature: str) -> Iterator[SignatureReturnType]:
        if signature == "0x":
            raise ValueError(f"Signature can not be: {signature}")

        data = self._get_all(
            endpoint=self.FUNCTION_ENDPOINT, filters={"hex_signature": signature}
        )

        for function in reversed(data):
            yield self._parse_text_signature_response(function)

    def get_event(self, signature: str) -> Iterator[SignatureReturnType]:
        if signature == "0x":
            raise ValueError(f"Signature can not be: {signature}")

        data = self._get_all(
            endpoint=self.EVENT_ENDPOINT, filters={"hex_signature": signature}
        )

        for event in reversed(data):
            yield self._parse_text_signature_response(event)

    def url(self, endpoint: str) -> str:
        return f"{self.API_URL}/{endpoint}/"

    def _get_all(self, endpoint: str, filters: Dict = None) -> List[Dict]:
        page = 1
        results = []

        while True:
            res = self._get(endpoint, page, filters)
            next_url = res.get("next")
            results.extend(res.get("results", []))

            if not next_url:
                break
            page += 1

        return results

    def _get(
        self, endpoint: str, page: int = 0, filters: Dict = None
    ) -> Dict[str, Any]:
        if filters is None:
            filters = {}

        if page:
            filters["page"] = page

        try:
            try:
                return requests.get(
                    self.url(endpoint), params=filters, timeout=5
                ).json()
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
            ) as e:
                raise FourByteConnectionException(e) from e
        except FourByteConnectionException:
            return {}

    @staticmethod
    def _parse_text_signature_response(data: Dict) -> SignatureReturnType:
        text_sig = data.get("text_signature", "")

        name = text_sig.split("(")[0] if text_sig else ""

        types = (
            text_sig[text_sig.find("(") + 1 : text_sig.rfind(")")] if text_sig else ""
        )
        if "(" in types:
            args = tuple(types[types.find("(") + 1 : types.rfind(")")].split(","))
        else:
            args = types.split(",")

        return {"name": name, "args": args}


FourByteProvider = FourByteProvider()



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from collections import OrderedDict
from typing import Dict, Optional

import requests

from ethtx.exceptions import ProcessingException

log = logging.getLogger(__name__)


class EtherscanClient:
    MODULE = "module="
    ACTION = "&action="
    CONTRACT_ADDRESS = "&contractaddress="
    ADDRESS = "&address="
    OFFSET = "&offset="
    PAGE = "&page="
    SORT = "&sort="
    BLOCK_TYPE = "&blocktype="
    TO = "&to="
    VALUE = "&value="
    DATA = "&data="
    POSITION = "&position="
    HEX = "&hex="
    GAS_PRICE = "&gasPrice="
    GAS = "&gas="
    START_BLOCK = "&startblock="
    END_BLOCK = "&endblock="
    BLOCKNO = "&blockno="
    TXHASH = "&txhash="
    TAG = "&tag="
    BOOLEAN = "&boolean="
    INDEX = "&index="
    API_KEY = "&apikey="

    url_dict: OrderedDict = {}

    def __init__(
        self,
        api_key: str,
        nodes: Dict[str, str],
        default_chain_id: Optional[str] = None,
    ):
        self.api_key = api_key
        self.endpoints = nodes
        self.default_chain = default_chain_id

        self.http = requests.session()
        self.http.headers.update({"User-Agent": "API"})

        self.url_dict = OrderedDict(
            [
                (self.MODULE, ""),
                (self.ADDRESS, ""),
                (self.ACTION, ""),
                (self.OFFSET, ""),
                (self.PAGE, ""),
                (self.SORT, ""),
                (self.BLOCK_TYPE, ""),
                (self.TO, ""),
                (self.VALUE, ""),
                (self.DATA, ""),
                (self.POSITION, ""),
                (self.HEX, ""),
                (self.GAS_PRICE, ""),
                (self.GAS, ""),
                (self.START_BLOCK, ""),
                (self.END_BLOCK, ""),
                (self.BLOCKNO, ""),
                (self.TXHASH, ""),
                (self.TAG, ""),
                (self.BOOLEAN, ""),
                (self.INDEX, ""),
                (self.API_KEY, api_key),
            ]
        )

    def build_url(self, chain_id: str, url_dict: OrderedDict) -> str:
        return (
            self.endpoints[chain_id]
            + "?"
            + "".join([param + val if val else "" for param, val in url_dict.items()])
        )

    def _get_chain_id(self, chain_id) -> str:
        _id = chain_id or self.default_chain

        if _id is None:
            raise ProcessingException(
                "chain_id must be provided as argument or constructor default"
            )
        return _id



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

from ethtx.models.decoded_model import DecodedBalance, DecodedTransactionMetadata
from .abc import SemanticSubmoduleAbc
from .helpers.utils import get_badge


class SemanticBalancesDecoder(SemanticSubmoduleAbc):
    """Semantic Balances Decoder."""

    def decode(
        self, balances: List[DecodedBalance], tx_metadata: DecodedTransactionMetadata
    ) -> List[DecodedBalance]:
        """Semantically decode balances."""

        for balance in balances:

            # decode the proper holder badge
            balance.holder.badge = get_badge(
                balance.holder.address,
                tx_metadata.sender.address,
                tx_metadata.receiver.address,
            )

            for token in balance.tokens:
                token["balance"] = f"{token['balance']:,.4f}"

        return balances



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import TransformationSemantics
from ethtx.semantics.base import Base


class Vat(Base):
    code_hash = "0x808b98f6475736d56c978e4fb476175ecd9d7abdab0797017fc10c7f46311a59"
    contract_semantics = dict(
        transformations={
            "0x2424be5c": {  # urns
                "art": TransformationSemantics(transformation="art / 10**18"),
                "ilk": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(ilk)"
                ),
            },
            "ink": TransformationSemantics(transformation="ink / 10**18"),
            "0x3b663195": {  # init
                "ilk": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(ilk)"
                )
            },
            "0x6111be2e": {  # flux
                "ilk": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(ilk)"
                ),
                "wad": TransformationSemantics(transformation="wad / 10**18"),
            },
            "0x7cdd3fde": {  # slip
                "ilk": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(ilk)"
                ),
                "wad": TransformationSemantics(transformation="wad / 10**18"),
            },
            "0x76088703": {  # frob
                "dart": TransformationSemantics(transformation="dart / 10**18"),
                "dink": TransformationSemantics(transformation="dink / 10**18"),
                "i": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(i)"
                ),
            },
            "0x870c616d": {  # fork
                "dart": TransformationSemantics(transformation="dart / 10**18"),
                "dink": TransformationSemantics(transformation="dink / 10**18"),
                "i": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(i)"
                ),
            },
            "0xb65337df": {  # fold
                "i": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(i)"
                ),
                "rate": TransformationSemantics(transformation="rate / 10**27"),
            },
            "0xbb35783b": {  # move
                "rad": TransformationSemantics(transformation="rad / 10**45")
            },
            "0xd9638d36": {  # ilks
                "__input0__": TransformationSemantics(
                    transformed_type="string",
                    transformation="string_from_bytes(__input0__)",
                ),
                "Art": TransformationSemantics(transformation="Art / 10**18"),
                "dust": TransformationSemantics(transformation="dust / 10**45"),
                "line": TransformationSemantics(transformation="line / 10**45"),
                "rate": TransformationSemantics(transformation="rate / 10**27"),
                "spot": TransformationSemantics(transformation="spot / 10**27"),
            },
        }
    )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from ethtx.models.semantics_model import FunctionSemantics, ParameterSemantics

precompiles = {
    1: FunctionSemantics(
        "",
        "ecrecover",
        [
            ParameterSemantics("hash", "bytes32", []),
            ParameterSemantics("v", "bytes8", []),
            ParameterSemantics("r", "bytes32", []),
            ParameterSemantics("s", "bytes32", []),
        ],
        [ParameterSemantics("", "address", [])],
    ),
    2: FunctionSemantics(
        "",
        "sha256",
        [ParameterSemantics("data", "raw", [])],
        [ParameterSemantics("", "bytes32", [])],
    ),
    3: FunctionSemantics(
        "",
        "ripemd160",
        [ParameterSemantics("data", "raw", [])],
        [ParameterSemantics("", "bytes32", [])],
    ),
    4: FunctionSemantics(
        "",
        "datacopy",
        [ParameterSemantics("data", "raw", [])],
        [ParameterSemantics("", "raw", [])],
    ),
    5: FunctionSemantics(
        "",
        "bigModExp",
        [
            ParameterSemantics("base", "bytes32", []),
            ParameterSemantics("exp", "bytes32", []),
            ParameterSemantics("mod", "bytes32", []),
        ],
        [ParameterSemantics("", "bytes32", [])],
    ),
    6: FunctionSemantics(
        "",
        "bn256Add",
        [
            ParameterSemantics("ax", "bytes32", []),
            ParameterSemantics("ay", "bytes32", []),
            ParameterSemantics("bx", "bytes32", []),
            ParameterSemantics("by", "bytes32", []),
        ],
        [ParameterSemantics("", "bytes32[2]", [])],
    ),
    7: FunctionSemantics(
        "",
        "bn256ScalarMul",
        [
            ParameterSemantics("x", "bytes32", []),
            ParameterSemantics("y", "bytes32", []),
            ParameterSemantics("scalar", "bytes32", []),
        ],
        [ParameterSemantics("", "bytes32[2]", [])],
    ),
    8: FunctionSemantics(
        "",
        "bn256Pairing",
        [ParameterSemantics("input", "raw", [])],
        [ParameterSemantics("", "bytes32", [])],
    ),
}



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from typing import Iterator, Optional, List

from ethtx.models.semantics_model import (
    FunctionSemantics,
    ParameterSemantics,
    Signature,
    SignatureArg,
)
from ethtx.providers.semantic_providers import SemanticsRepository
from ethtx.providers.signature_provider import SignatureProvider, FourByteProvider

log = logging.getLogger(__name__)


def decode_function_abi_with_external_source(
    signature: str,
    repository: SemanticsRepository,
    _provider: Optional[SignatureProvider] = FourByteProvider,
) -> Iterator[FunctionSemantics]:
    function = repository.get_most_used_signature(signature_hash=signature)
    if function:
        log.debug(
            "Successfully guessed function from SemanticsRepository - %s.",
            function.json(),
        )
        function_semantics = FunctionSemantics(
            signature,
            function.name,
            _prepare_parameter_semantics(function.args, function.tuple, unknown=False),
            [],
        )
        yield function_semantics
        return

    functions = _provider.get_function(signature=signature)
    try:
        for func in functions:
            if not func:
                yield

            function_semantics = FunctionSemantics(
                signature,
                func.get("name"),
                _prepare_parameter_semantics(
                    func.get("args"), isinstance(func.get("args"), tuple), unknown=True
                ),
                [],
            )
            yield function_semantics
    finally:
        if "function_semantics" in locals():
            repository.update_or_insert_signature(
                signature=Signature(
                    signature_hash=signature,
                    name=function_semantics.name,
                    args=[
                        SignatureArg(name=f"arg_{i}", type=arg)
                        for i, arg in enumerate(func.get("args"))
                    ],
                    tuple=isinstance(func.get("args"), tuple),
                )
            )


def decode_event_abi_name_with_external_source(
    signature: str, _provider: Optional[SignatureProvider] = FourByteProvider
) -> str:
    events = _provider.get_event(signature=signature)

    for event in events:

        if not event:
            return signature

        return event.get("name", signature)

    return signature


def _prepare_parameter_semantics(
    args, is_tuple: bool, unknown: bool
) -> List[ParameterSemantics]:
    if not is_tuple:
        return [
            ParameterSemantics(
                arg["name"] if not unknown else f"arg_{i}",
                arg["type"] if not unknown else arg,
                [],
            )
            for i, arg in enumerate(args)
        ]

    return [
        ParameterSemantics(
            "params",
            "tuple",
            [
                ParameterSemantics(
                    arg["name"] if not unknown else f"arg_{i}",
                    arg["type"] if not unknown else arg,
                    [],
                )
                for i, arg in enumerate(args)
            ],
        )
    ]



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Dict, Union, Optional

from ethtx.models.decoded_model import DecodedEvent, Proxy
from ethtx.models.objects_model import BlockMetadata, TransactionMetadata, Event
from ethtx.semantics.standards.erc20 import ERC20_EVENTS
from ethtx.semantics.standards.erc721 import ERC721_EVENTS
from .abc import ABISubmoduleAbc
from .helpers.utils import decode_event_abi_name_with_external_source
from ..decoders.parameters import decode_event_parameters


class ABIEventsDecoder(ABISubmoduleAbc):
    """ABI Events Decoder."""

    def decode(
        self,
        events: Union[Event, List[Event]],
        block: BlockMetadata,
        transaction: TransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
        chain_id: Optional[str] = None,
    ) -> Union[DecodedEvent, List[DecodedEvent]]:
        """Return list of decoded events."""
        if isinstance(events, list):
            return (
                [
                    self.decode_event(event, block, transaction, proxies, chain_id)
                    for event in events
                ]
                if events
                else []
            )

        return self.decode_event(events, block, transaction, proxies, chain_id)

    def decode_event(
        self,
        event: Event,
        block: BlockMetadata,
        transaction: TransactionMetadata,
        proxies: Dict[str, Proxy] = None,
        chain_id: str = None,
    ) -> DecodedEvent:

        if event.topics:
            event_signature = event.topics[0]
        else:
            event_signature = None

        anonymous = False
        chain_id = chain_id or self._default_chain

        event_abi = self._repository.get_event_abi(
            chain_id, event.contract, event_signature
        )

        if not event_abi:

            if not event_abi:
                # if signature is not known but there is exactly one anonymous event in tha ABI
                # we can assume that this is this the anonymous one (e.g. Maker's LogNote)
                event_abi = self._repository.get_anonymous_event_abi(
                    chain_id, event.contract
                )
                if event_abi:
                    anonymous = True

            if not event_abi and event.contract in proxies:
                # try to find signature in delegate-called contracts
                for semantic in proxies[event.contract].semantics:
                    event_abi = (
                        semantic.contract.events[event_signature]
                        if event_signature in semantic.contract.events
                        else None
                    )
                    if event_abi:
                        break

            if not event_abi and event_signature in ERC20_EVENTS:
                # try standard ERC20 events
                if (
                    len(
                        [
                            parameter
                            for parameter in ERC20_EVENTS[event_signature].parameters
                            if parameter.indexed
                        ]
                    )
                    == len([topic for topic in event.topics if topic]) - 1
                ):
                    event_abi = ERC20_EVENTS[event_signature]
                elif event_signature in ERC721_EVENTS:
                    # try standard ERC721 events
                    if (
                        len(
                            [
                                parameter
                                for parameter in ERC721_EVENTS[
                                    event_signature
                                ].parameters
                                if parameter.indexed
                            ]
                        )
                        == len([topic for topic in event.topics if topic]) - 1
                    ):
                        event_abi = ERC721_EVENTS[event_signature]

        contract_name = self._repository.get_address_label(
            chain_id, event.contract, proxies
        )
        event_name = event_abi.name if event_abi else event_signature

        parameters = decode_event_parameters(
            event.log_data, event.topics, event_abi, anonymous
        )

        if event_name.startswith("0x") and len(event_name) > 2:
            event_name = decode_event_abi_name_with_external_source(
                signature=event_signature
            )

        return DecodedEvent(
            chain_id=chain_id,
            tx_hash=transaction.tx_hash,
            timestamp=block.timestamp,
            contract_address=event.contract,
            contract_name=contract_name,
            index=event.log_index,
            call_id=event.call_id,
            event_signature=event_signature,
            event_name=event_name,
            parameters=parameters,
        )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from typing import Dict, List, Union

from .abi.decoder import ABIDecoder
from .semantic.decoder import SemanticDecoder
from ..models.decoded_model import DecodedTransaction, Proxy
from ..models.objects_model import Block, Call
from ..providers.web3_provider import NodeDataProvider
from ..semantics.standards.eip1969 import is_eip1969_proxy, is_eip1969_beacon_proxy

log = logging.getLogger(__name__)


class DecoderService:
    def __init__(
        self,
        abi_decoder: ABIDecoder,
        semantic_decoder: SemanticDecoder,
        web3provider: NodeDataProvider,
        default_chain: str,
    ):
        self.abi_decoder: ABIDecoder = abi_decoder
        self.semantic_decoder: SemanticDecoder = semantic_decoder
        self.web3provider: NodeDataProvider = web3provider
        self.default_chain: str = default_chain

    def get_delegations(self, calls: Union[Call, List[Call]]) -> Dict[str, List[str]]:

        delegations = {}

        if not calls:
            return delegations

        if isinstance(calls, list):
            for call in calls:
                if call.call_type == "delegatecall":
                    if call.from_address not in delegations:
                        delegations[call.from_address] = []
                    if call.to_address not in delegations[call.from_address]:
                        delegations[call.from_address].append(call.to_address)
        else:
            calls_queue = [calls]

            while calls_queue:
                call = calls_queue.pop()
                for _, sub_call in enumerate(call.subcalls):
                    calls_queue.insert(0, sub_call)

                if call.call_type == "delegatecall":
                    if call.from_address not in delegations:
                        delegations[call.from_address] = []
                    if call.to_address not in delegations[call.from_address]:
                        delegations[call.from_address].append(call.to_address)

        return delegations

    def get_proxies(
        self, delegations: Dict[str, List[str]], chain_id: str
    ) -> Dict[str, Proxy]:

        proxies = {}
        chain = self.web3provider._get_node_connection(chain_id)

        for delegator in delegations:

            delegator_semantics = self.semantic_decoder.repository.get_semantics(
                chain_id, delegator
            )

            if is_eip1969_proxy(chain, delegator, delegations[delegator][0]):
                proxy_type = "EIP1969Proxy"
                fallback_name = "EIP1969_Proxy"

            elif is_eip1969_beacon_proxy(chain, delegator, delegations[delegator][0]):
                proxy_type = "EIP1969Beacon"
                fallback_name = "EIP1969_BeaconProxy"

            else:
                proxy_type = "GenericProxy"
                fallback_name = "Proxy"

            delegates_semantics = [
                self.semantic_decoder.repository.get_semantics(chain_id, delegate)
                for delegate in delegations[delegator]
            ]

            token_semantics = delegator_semantics.erc20
            if not token_semantics:
                for delegate_semantics in delegates_semantics:
                    if delegate_semantics.erc20:
                        token_semantics = delegate_semantics.erc20
                        break

            proxies[delegator] = Proxy(
                address=delegator,
                name=delegator_semantics.name
                if delegator_semantics and delegator_semantics.name != delegator
                else fallback_name,
                type=proxy_type,
                semantics=[semantics for semantics in delegates_semantics if semantics],
                token=token_semantics,
            )

        return proxies

    def decode_transaction(self, chain_id: str, tx_hash: str) -> DecodedTransaction:

        # verify the transaction hash
        tx_hash = tx_hash if tx_hash.startswith("0x") else "0x" + tx_hash

        chain_id = chain_id or self.default_chain

        self.semantic_decoder.repository.record()
        # read a raw transaction from a node
        transaction = self.web3provider.get_full_transaction(
            tx_hash=tx_hash, chain_id=chain_id
        )
        # read a raw block from a node
        block = Block.from_raw(
            w3block=self.web3provider.get_block(
                transaction.metadata.block_number, chain_id
            ),
            chain_id=chain_id,
        )

        # prepare lists of delegations to properly decode delegate-calling contracts
        delegations = self.get_delegations(transaction.root_call)
        proxies = self.get_proxies(delegations, chain_id)

        # decode transaction using ABI
        abi_decoded_tx = self.abi_decoder.decode_transaction(
            block=block, transaction=transaction, proxies=proxies, chain_id=chain_id
        )

        # decode transaction using additional semantics
        semantically_decoded_tx = self.semantic_decoder.decode_transaction(
            block=block.metadata,
            transaction=abi_decoded_tx,
            proxies=proxies,
            chain_id=chain_id,
        )

        used_semantics = self.semantic_decoder.repository.end_record()
        log.info(
            "Semantics used in decoding %s: %s",
            tx_hash,
            ", ".join(used_semantics) if used_semantics else "",
        )

        return semantically_decoded_tx



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from abc import ABC, abstractmethod
from typing import Callable, Any

from ens import ENS
from eth_typing import ChecksumAddress
from web3 import Web3


class ENSProviderABC(ABC):
    @abstractmethod
    def __init__(self, provider: Callable):
        ...

    @abstractmethod
    def name(self, address: Any):
        ...

    @abstractmethod
    def address(self, name: Any):
        ...


class Web3ENSProvider(ENSProviderABC):
    ns: ENS

    def __init__(self, provider: Web3):
        self.ns = ENS.fromWeb3(provider)

    def name(self, address: ChecksumAddress) -> str:
        return self.ns.name(address=address)

    def address(self, name: str) -> str:
        return self.ns.address(name=name)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import TypeVar, Dict


class Base(ABC):
    """
    Semantics Base class.
    All semantics should have `code_hash: str` and `contract_semantics: dict` class attribute.
    Inherit this class if you want to implement new semantic.
    """

    @property
    @abstractmethod
    def contract_semantics(self) -> str:
        ...

    @property
    @abstractmethod
    def code_hash(self) -> dict:
        ...


BaseType = TypeVar(
    "BaseType", bound=Dict[Base.code_hash.fget, Base.contract_semantics.fget]
)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import time
from functools import partial
from typing import Optional

import requests

from ethtx.decoders.decoders.parameters import decode_function_parameters
from ethtx.models.decoded_model import AddressInfo
from ethtx.models.semantics_model import FunctionSemantics
from ethtx.semantics.utilities.functions import add_utils_to_context

log = logging.getLogger(__name__)

eth_price: Optional[float] = None
eth_price_update: Optional[float] = None


def get_eth_price() -> Optional[float]:
    global eth_price, eth_price_update

    current_time = time.time()
    if (
        eth_price is None
        or eth_price_update is None
        or (current_time - eth_price_update) > 60
    ):
        response = requests.get("https://api.coinbase.com/v2/prices/ETH-USD/buy")
        if response.status_code == 200:
            eth_price = float(json.loads(response.content)["data"]["amount"])
            eth_price_update = time.time()

    return eth_price


def get_badge(address, sender, receiver):
    sender_address = sender.address if isinstance(sender, AddressInfo) else sender
    receiver_address = (
        receiver.address if isinstance(receiver, AddressInfo) else receiver
    )

    if address == sender_address:
        badge = "sender"
    elif address == receiver_address:
        badge = "receiver"
    else:
        badge = None

    return badge


def semantically_decode_parameter(
    repository, parameter, indexed_name, transformations, proxies, context
):
    if parameter.name in transformations:
        transformation = transformations[parameter.name]
    elif indexed_name in transformations:
        transformation = transformations[indexed_name]
    else:
        transformation = None

    if transformation:
        parameter.name = transformation.transformed_name or parameter.name
        parameter.type = transformation.transformed_type or parameter.type

        if transformation.transformation:
            parameter.value = (
                evaluate_transformation(
                    parameter.value, transformation.transformation, context
                )
                or parameter.value
            )

    if parameter.type == "address" and not isinstance(parameter.value, AddressInfo):
        address = parameter.value
        name = repository.get_address_label(
            context["__transaction__"].chain_id, address, proxies
        )
        badge = get_badge(
            address,
            context["__transaction__"].sender,
            context["__transaction__"].receiver,
        )
        parameter.value = AddressInfo(address, name, badge)
    elif parameter.type == "bytes" and isinstance(parameter.value, str):
        if len(parameter.value) > 66:
            parameter.value = parameter.value[:60] + "..." + parameter.value[-6:]
    elif parameter.type == "tuple" and isinstance(parameter.value, list):
        for i, sub_parameter in enumerate(parameter.value):
            semantically_decode_parameter(
                repository,
                sub_parameter,
                f"__input{i}__",
                transformations,
                proxies,
                context,
            )


def evaluate_transformation(value, transformation, context):
    try:
        new_value = eval(transformation, context)
    except Exception as e:
        log.warning("Transformation: %s failed.", transformation, exc_info=e)
        new_value = value

    return new_value


def decode_call(transaction, repository, contract_address, data):
    if not data or len(data) <= 2:
        return None

    function_signature = data[:10]

    contract_name = repository.get_address_label(transaction.chain_id, contract_address)
    contract_badge = get_badge(
        contract_address, transaction.sender, transaction.receiver
    )
    contract = AddressInfo(contract_address, contract_name, contract_badge)

    if repository.check_is_contract(transaction.chain_id, contract_address):

        # read function ABI and semantics and use them for decoding
        function_abi = repository.get_function_abi(
            transaction.chain_id, contract_address, function_signature
        )
        function_transformations = repository.get_transformations(
            transaction.chain_id, contract_address, function_signature
        )
        function_name = function_abi.name if function_abi else function_signature
        stripped_function_abi = FunctionSemantics(
            signature=function_abi.signature,
            name=function_abi.name,
            inputs=function_abi.inputs,
            outputs=[],
        )
        function_input, _ = decode_function_parameters(
            data, "0x", stripped_function_abi
        )

        # perform arguments transformations
        context = create_transformation_context(
            contract.address, function_input, [], transaction, repository
        )
        for i, argument in enumerate(function_input):
            semantically_decode_parameter(
                repository,
                argument,
                f"__input{i}__",
                function_transformations or {},
                None,
                context,
            )

        decoded_argument = dict(
            contract=contract, function_name=function_name, arguments=function_input
        )
    else:
        decoded_argument = dict(
            contract=contract,
            function_name=function_signature,
            arguments="0x" + data[10:],
        )

    return decoded_argument


def create_transformation_context(
    contract, input_variables, output_variables, transaction, repository
):
    # create a context for transformations
    context = {}
    for i, parameter in enumerate(input_variables):
        if parameter.name:
            context[parameter.name] = parameter.value
        context[f"__input{i}__"] = parameter.value

    for i, parameter in enumerate(output_variables):
        if parameter.name:
            context[parameter.name] = parameter.value
        context[f"__output{i}__"] = parameter.value

    # register context variables
    context["__transaction__"] = transaction
    context["__contract__"] = contract
    context["__repository__"] = repository

    # register additional functions available for transformations
    context["decode_call"] = partial(
        decode_call, context["__transaction__"], context["__repository__"]
    )
    add_utils_to_context(context)

    # ToDo: remove compatibility hooks
    # compatibility hooks
    context["_transaction"] = transaction
    context["_contract"] = contract
    context["network"] = transaction.chain_id
    context["_print_input"] = decode_call
    context["__print_input__"] = decode_call

    return context



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


def decode_rollup_data(data):
    def get_32word_at(data, pos):
        word_length = 2 * 32
        dataStart = pos * 2
        return data[dataStart : dataStart + word_length]

    numberOfAssets = 4
    txNumPubInputs = 12
    rollupNumPubInputs = 10 + numberOfAssets
    # public inputs length for of each inner proof tx
    txPubInputLength = txNumPubInputs * 32
    rollupPubInputLength = rollupNumPubInputs * 32

    rollup = dict(
        rollupPubInputLength=rollupPubInputLength,
        rollupId=int(get_32word_at(data, 0), 16),
        rollupSize=int(get_32word_at(data, int("0x20", 16)), 16),
        dataStartIndex=int(get_32word_at(data, int("0x40", 16)), 16),
        oldDataRoot=get_32word_at(data, int("0x60", 16)),
        newDataRoot=get_32word_at(data, int("0x80", 16)),
        oldNullRoot=get_32word_at(data, int("0xa0", 16)),
        newNullRoot=get_32word_at(data, int("0xc0", 16)),
        oldRootRoot=get_32word_at(data, int("0xe0", 16)),
        newRootRoot=get_32word_at(data, int("0x100", 16)),
    )

    rollupSize = int(get_32word_at(data, int("0x20", 16)), 16)
    numTxs = 1 if rollupSize == 0 else rollupSize

    proofDataPointer = rollupPubInputLength
    operations = []
    for _ in range(numTxs):

        proofId = int(get_32word_at(data, proofDataPointer), 16)
        publicInput = get_32word_at(data, proofDataPointer + int("0x20", 16))
        publicOutput = get_32word_at(data, proofDataPointer + int("0x40", 16))
        nullifier1 = get_32word_at(data, proofDataPointer + int("0x100", 16))

        if proofId == 0:

            assetId = int(get_32word_at(data, proofDataPointer + int("0x60", 16)), 16)
            inputOwner = (
                "0x" + get_32word_at(data, proofDataPointer + int("0x140", 16))[-40:]
            )
            outputOwner = (
                "0x" + get_32word_at(data, proofDataPointer + int("0x160", 16))[-40:]
            )

            if assetId == 0:
                if int(publicInput, 16):
                    operation = dict(
                        type="Deposits",
                        address=inputOwner,
                        amount=f"{int(publicInput, 16) / 10 ** 18:,} ETH",
                    )
                    operations.append(operation)

                elif int(publicOutput, 16):
                    operation = dict(
                        type="Withdrawals",
                        address=outputOwner,
                        amount=f"{int(publicOutput, 16) / 10 ** 18:,} ETH",
                    )
                    operations.append(operation)

                elif int(nullifier1, 16):
                    operation = dict(type="Private", address="", amount="")
                    operations.append(operation)

        elif proofId == 1:

            address = "0x" + publicInput + publicOutput
            operation = dict(
                type="Accounts", address=address[:60] + "..." + address[-6:], amount=""
            )
            operations.append(operation)

        proofDataPointer = proofDataPointer + txPubInputLength

    return rollup, operations



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

from ethtx.models.decoded_model import DecodedCall, DecodedTransfer, AddressInfo
from ethtx.utils.measurable import RecursionLimit
from .abc import ABISubmoduleAbc

RECURSION_LIMIT = 2000
ZERO_ADDRESS = "0x" + 40 * "0"


class ABITransfersDecoder(ABISubmoduleAbc):
    """Abi Transfers Decoder."""

    def decode(self, call: DecodedCall, events, proxies) -> List:
        """Decode transfers."""
        transfers = []

        def _transfers_calls(decoded_call):
            if decoded_call.status and decoded_call.value:
                transfers.append(
                    DecodedTransfer(
                        from_address=decoded_call.from_address,
                        to_address=decoded_call.to_address,
                        token_standard="ETH",
                        token_address=ZERO_ADDRESS,
                        token_symbol="ETH",
                        value=decoded_call.value,
                    )
                )
            if decoded_call.subcalls:
                for sub_call in decoded_call.subcalls:
                    _transfers_calls(sub_call)

        if call:
            with RecursionLimit(RECURSION_LIMIT):
                _transfers_calls(call)

        for event in events:

            if event.event_name == "Transfer":

                from_address = event.parameters[0].value
                from_name = self._repository.get_address_label(
                    event.chain_id, from_address, proxies
                )
                to_address = event.parameters[1].value
                to_name = self._repository.get_address_label(
                    event.chain_id, to_address, proxies
                )

                standard = self._repository.get_standard(
                    event.chain_id, event.contract.address
                )

                if standard == "ERC20" or event.contract.address in proxies:

                    (
                        _,
                        token_symbol,
                        token_decimals,
                        _,
                    ) = self._repository.get_token_data(
                        event.chain_id, event.contract.address, proxies
                    )
                    value = event.parameters[2].value / 10 ** token_decimals
                    transfers.append(
                        DecodedTransfer(
                            from_address=AddressInfo(from_address, from_name),
                            to_address=AddressInfo(to_address, to_name),
                            token_standard=standard,
                            token_address=event.contract.address,
                            token_symbol=token_symbol,
                            value=value,
                        )
                    )
                elif standard == "ERC721":
                    if len(str(event.parameters[2].value)) > 8:
                        token_symbol = (
                            f"NFT {str(event.parameters[2].value)[:6]}..."
                            f"{str(event.parameters[2].value)[-2:]}"
                        )
                    else:
                        token_symbol = f"NFT {event.parameters[2].value}"
                    token_address = f"{event.contract.address}?a={event.parameters[2].value}#inventory"
                    transfers.append(
                        DecodedTransfer(
                            from_address=AddressInfo(from_address, from_name),
                            to_address=AddressInfo(to_address, to_name),
                            token_standard=standard,
                            token_address=token_address,
                            token_symbol=token_symbol,
                            value=1,
                        )
                    )

        return transfers



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import sys
import time

log = logging.getLogger(__name__)


class ExecutionTimer:
    start_time: float
    part_name: str

    def __init__(self, part_name: str):
        self.part_name = part_name

    def __enter__(self):
        self.start_time = time.time()

    def __exit__(self, *kwargs):
        end_time = time.time()
        exec_time = (end_time - self.start_time) * 1000
        log.info("Executed %s in %s ms", self.part_name, exec_time)


class RecursionLimit:
    def __init__(self, limit: int):
        self.limit = limit
        self.cur_limit = sys.getrecursionlimit()

    def __enter__(self) -> None:
        sys.setrecursionlimit(self.limit)

    def __exit__(self, *_) -> None:
        sys.setrecursionlimit(self.cur_limit)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Dict, Optional

from .contracts import EtherscanContract


class EtherscanProvider:
    api_key: str
    endpoints: Dict[str, str]
    default_chain: Optional[str]

    def __init__(
        self,
        api_key: str,
        nodes: Dict[str, str],
        default_chain_id: Optional[str] = None,
    ):
        self.api_key = api_key
        self.endpoints = nodes
        self.default_chain = default_chain_id

        self._contract = EtherscanContract(
            api_key=self.api_key,
            nodes=self.endpoints,
            default_chain_id=self.default_chain,
        )

    @property
    def contract(self) -> EtherscanContract:
        return self._contract



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import TransformationSemantics
from ethtx.semantics.base import Base


class Dai(Base):
    code_hash = "0x4e36f96ee1667a663dfaac57c4d185a0e369a3a217e0079d49620f34f85d1ac7"
    contract_semantics = dict(
        transformations={
            "0x40c10f19": {  # mint
                "wad": TransformationSemantics(transformation="wad / 10**18")
            },
            "0x9dc29fac": {  # burn
                "wad": TransformationSemantics(transformation="wad / 10**18")
            },
        }
    )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Dict, List, Optional

from ethtx.decoders.semantic.abc import ISemanticDecoder
from ethtx.decoders.semantic.balances import SemanticBalancesDecoder
from ethtx.decoders.semantic.calls import SemanticCallsDecoder
from ethtx.decoders.semantic.events import SemanticEventsDecoder
from ethtx.decoders.semantic.metadata import SemanticMetadataDecoder
from ethtx.decoders.semantic.transfers import SemanticTransfersDecoder
from ethtx.models.decoded_model import (
    DecodedTransactionMetadata,
    DecodedTransaction,
    DecodedEvent,
    DecodedTransfer,
    DecodedBalance,
    DecodedCall,
    Proxy,
)
from ethtx.models.objects_model import BlockMetadata, TransactionMetadata


class SemanticDecoder(ISemanticDecoder):
    def decode_transaction(
        self,
        block: BlockMetadata,
        transaction: DecodedTransaction,
        proxies: Dict[str, Proxy],
        chain_id: str,
    ) -> DecodedTransaction:
        transaction.metadata = self.decode_metadata(
            block, transaction.metadata, chain_id
        )
        transaction.events = self.decode_events(
            transaction.events, transaction.metadata, proxies
        )
        transaction.calls = self.decode_calls(
            transaction.calls, transaction.metadata, proxies
        )
        transaction.transfers = self.decode_transfers(
            transaction.transfers, transaction.metadata
        )
        transaction.balances = self.decode_balances(
            transaction.balances, transaction.metadata
        )

        return transaction

    def decode_metadata(
        self,
        block_metadata: BlockMetadata,
        tx_metadata: TransactionMetadata,
        chain_id: str,
    ) -> DecodedTransactionMetadata:
        return SemanticMetadataDecoder(repository=self.repository).decode(
            block_metadata=block_metadata, tx_metadata=tx_metadata, chain_id=chain_id
        )

    def decode_event(
        self,
        event: DecodedEvent,
        tx_metadata: DecodedTransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
    ) -> DecodedEvent:
        return SemanticEventsDecoder(repository=self.repository).decode(
            events=event, tx_metadata=tx_metadata, proxies=proxies or {}
        )

    def decode_events(
        self,
        events: List[DecodedEvent],
        tx_metadata: DecodedTransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
    ) -> List[DecodedEvent]:
        return SemanticEventsDecoder(repository=self.repository).decode(
            events=events, tx_metadata=tx_metadata, proxies=proxies or {}
        )

    def decode_calls(
        self,
        call: DecodedCall,
        tx_metadata: DecodedTransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
    ) -> DecodedCall:
        return SemanticCallsDecoder(repository=self.repository).decode(
            call=call, tx_metadata=tx_metadata, proxies=proxies or {}
        )

    def decode_call(
        self,
        call: DecodedCall,
        tx_metadata: DecodedTransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
    ) -> DecodedCall:
        return SemanticCallsDecoder(repository=self.repository).decode(
            call=call, tx_metadata=tx_metadata, proxies=proxies or {}
        )

    def decode_transfers(
        self, transfers: List[DecodedTransfer], tx_metadata: DecodedTransactionMetadata
    ) -> List[DecodedTransfer]:
        return SemanticTransfersDecoder(repository=self.repository).decode(
            transfers=transfers, tx_metadata=tx_metadata
        )

    def decode_balances(
        self, balances: List[DecodedBalance], tx_metadata: DecodedTransactionMetadata
    ) -> List[DecodedBalance]:
        return SemanticBalancesDecoder(repository=self.repository).decode(
            balances=balances, tx_metadata=tx_metadata
        )



from web3 import Web3


def is_eip1969_proxy(chain, delegator, delegate):
    implementation_slot = hex(
        int(Web3.keccak(text="eip1967.proxy.implementation").hex(), 16) - 1
    )
    try:
        implementation = (
            "0x"
            + chain.eth.get_storage_at(
                Web3.toChecksumAddress(delegator), implementation_slot
            ).hex()[-40:]
        )
        return implementation == delegate
    except:
        return False


def is_eip1969_beacon_proxy(chain, delegator, delegate):
    ibeacon_abi = """[
                        {
                            "inputs": [],
                            "name": "implementation",
                            "outputs": [
                                {
                                    "internalType": "address",
                                    "name": "",
                                    "type": "address"
                                }
                            ],
                            "stateMutability": "view",
                            "type": "function"
                        }
                    ]"""

    beacon_slot = hex(int(Web3.keccak(text="eip1967.proxy.beacon").hex(), 16) - 1)
    try:
        beacon = (
            "0x"
            + chain.eth.get_storage_at(
                Web3.toChecksumAddress(delegator), beacon_slot
            ).hex()[-40:]
        )
        beacon = chain.eth.contract(
            address=Web3.toChecksumAddress(beacon), abi=ibeacon_abi
        )
        implementation = beacon.functions.implementation().call()
        return implementation == Web3.toChecksumAddress(delegate)
    except:
        return False



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import TransformationSemantics
from ethtx.semantics.base import Base


class Jug(Base):
    code_hash = "0x79e3139c72f7ad381e81a7ff8ae63b5998f0778c1c025c6095a6f878771dc729"
    contract_semantics = dict(
        transformations={
            "0x44e2a5a8": {  # drip
                "ilk": TransformationSemantics(
                    transformed_type="string", transformation="string_from_bytes(ilk)"
                )
            }
        }
    )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC
from typing import Dict, Optional, Any, List


class ISemanticsDatabase(ABC):
    """Semantics Database. Represents raw interface required to be
    implemented by a database that provides persistent
    data about semantics"""

    def get_address_semantics(self, chain_id: str, address: str) -> Optional[Dict]:
        ...

    def get_contract_semantics(self, code_hash: str) -> Optional[Dict]:
        ...

    def get_signature_semantics(self, signature_hash: str) -> Optional[List[Dict]]:
        ...

    def insert_contract(self, contract: dict, update_if_exist: bool = False) -> Any:
        ...

    def insert_address(self, address: dict, update_if_exist: bool = False) -> Any:
        ...

    def insert_signature(self, signature, update_if_exist: bool = False) -> Any:
        ...



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Dict, Optional

import bson
from pymongo import TEXT
from pymongo.cursor import Cursor
from pymongo.database import Database as MongoDatabase
from pymongo.errors import OperationFailure

from .base import ISemanticsDatabase
from .const import MongoCollections

log = logging.getLogger(__name__)


class MongoSemanticsDatabase(ISemanticsDatabase):
    _db: MongoDatabase

    def __init__(self, db: MongoDatabase):
        self._db = db

        self._addresses = None
        self._contracts = None
        self._signatures = None

        self._init_collections()

    def get_collection_count(self) -> int:
        return len(self._db.list_collection_names())

    def get_address_semantics(self, chain_id, address) -> Optional[Dict]:
        _id = f"{chain_id}-{address}"
        return self._addresses.find_one({"_id": _id}, {"_id": 0})

    def get_signature_semantics(self, signature_hash: str) -> Cursor:
        return self._signatures.find({"signature_hash": signature_hash})

    def insert_signature(
        self, signature: dict, update_if_exist=False
    ) -> Optional[bson.ObjectId]:
        if update_if_exist:
            updated_signature = self._signatures.replace_one(
                {"_id": signature["_id"]}, signature, upsert=True
            )
            return (
                None
                if updated_signature.modified_count
                else updated_signature.upserted_id
            )

        inserted_signature = self._signatures.insert_one(signature)
        return inserted_signature.inserted_id

    def get_contract_semantics(self, code_hash):
        """Contract hashes are always the same, no mather what chain we use, so there is no need
        to use chain_id"""
        return self._contracts.find_one({"_id": code_hash}, {"_id": 0})

    def insert_contract(
        self, contract, update_if_exist=False
    ) -> Optional[bson.ObjectId]:
        contract_with_id = {"_id": contract["code_hash"], **contract}

        if update_if_exist:
            updated_contract = self._contracts.replace_one(
                {"_id": contract_with_id["_id"]}, contract_with_id, upsert=True
            )

            return (
                None
                if updated_contract.modified_count
                else updated_contract.upserted_id
            )

        inserted_contract = self._contracts.insert_one(contract_with_id)
        return inserted_contract.inserted_id

    def insert_address(self, address, update_if_exist=False) -> Optional[bson.ObjectId]:
        address_with_id = {
            "_id": f"{address['chain_id']}-{address['address']}",
            **address,
        }

        if update_if_exist:
            updated_address = self._addresses.replace_one(
                {"_id": address_with_id["_id"]}, address_with_id, upsert=True
            )
            return (
                None if updated_address.modified_count else updated_address.upserted_id
            )

        inserted_address = self._addresses.insert_one(address_with_id)
        return inserted_address.inserted_id

    def _init_collections(self) -> None:
        for mongo_collection in MongoCollections:
            self.__setattr__(f"_{mongo_collection}", self._db[mongo_collection])

            if mongo_collection == "signatures":
                try:
                    self._signatures.create_index(
                        [("signature_hash", TEXT), ("name", TEXT)],
                        background=True,
                        unique=False,
                    )
                except OperationFailure as e:
                    log.warning(e)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
from functools import lru_cache
from typing import Dict, Tuple, Union, Any, Optional

from web3 import Web3

from ethtx.exceptions import InvalidEtherscanReturnCodeException
from .client import EtherscanClient

log = logging.getLogger(__name__)


class EtherscanContract(EtherscanClient):
    def __init__(
        self,
        api_key: str,
        nodes: Dict[str, str],
        default_chain_id: Optional[str] = None,
    ):
        EtherscanClient.__init__(
            self, api_key=api_key, nodes=nodes, default_chain_id=default_chain_id
        )
        self.contract_dict = self.url_dict.copy()
        self.contract_dict[self.MODULE] = "contract"

    def get_contract_abi(
        self, chain_id, contract_name
    ) -> Tuple[Dict[str, Union[dict, Any]], bool]:
        decoded = False
        raw_abi = []

        try:
            resp = self._get_contract_abi(chain_id, contract_name)
            if resp["status"] == "1" and resp["message"] == "OK":
                contract_name = resp["result"][0]["ContractName"]
                if (
                    len(resp["result"][0]["ABI"])
                    and resp["result"][0]["ABI"] != "Contract source code not verified"
                ):
                    raw_abi = json.loads(resp["result"][0]["ABI"])
                    decoded = True

        except Exception as e:
            log.exception(
                "Etherscan connection failed while getting abi for %s on %s",
                contract_name,
                chain_id,
                exc_info=e,
            )

        abi = self._parse_abi(raw_abi)

        return dict(name=contract_name, abi=abi), decoded

    @lru_cache(maxsize=1024)
    def _get_contract_abi(self, chain_id, contract_name) -> Dict:
        url_dict = self.contract_dict.copy()
        url_dict[self.ACTION] = "getsourcecode"
        url_dict[self.ADDRESS] = contract_name
        url = self.build_url(chain_id=self._get_chain_id(chain_id), url_dict=url_dict)

        # TODO: etherscan sometimes returns HTTP 502 with no apparent reason, so it's a quick fix
        # that should help, but number of tries should be taken from config in final solution I think
        for _ in range(3):
            resp = self.http.get(url)

            if resp.status_code == 200:
                break

        if resp.status_code != 200:
            raise InvalidEtherscanReturnCodeException(resp.status_code, url_dict)

        return resp.json()

    # helper function decoding contract ABI
    @staticmethod
    def _parse_abi(json_abi) -> Dict:
        # helper function to recursively parse components
        def _parse_components(components):

            comp_canonical = "("
            comp_inputs = []

            for i, component in enumerate(components):

                argument = dict(name=component["name"], type=component["type"])

                if component["type"][:5] == "tuple":
                    sub_canonical, sub_components = _parse_components(
                        component["components"]
                    )
                    comp_canonical += sub_canonical + component["type"][5:]
                    argument["components"] = sub_components
                else:
                    comp_canonical += component["type"]
                    sub_components = []

                if i < len(components) - 1:
                    comp_canonical += ","

                if (
                    component["type"] in ("string", "bytes")
                    or component["type"][-2:] == "[]"
                ):
                    argument["dynamic"] = True
                elif component["type"] == "tuple":
                    argument["dynamic"] = any(c["dynamic"] for c in sub_components)
                else:
                    argument["dynamic"] = False

                if "indexed" in component:
                    argument["indexed"] = component["indexed"]

                comp_inputs.append(argument)

            comp_canonical += ")"

            return comp_canonical, comp_inputs

        functions = {}
        events = {}

        for item in json_abi:

            if "type" in item:

                # parse contract functions
                if item["type"] == "constructor":
                    _, inputs = _parse_components(item["inputs"])
                    functions["constructor"] = dict(
                        signature="constructor",
                        name="constructor",
                        inputs=inputs,
                        outputs=[],
                    )

                elif item["type"] == "fallback":
                    functions["fallback"] = {}

                elif item["type"] == "function":
                    canonical, inputs = _parse_components(item["inputs"])
                    canonical = item["name"] + canonical
                    function_hash = Web3.sha3(text=canonical).hex()
                    signature = function_hash[0:10]

                    _, outputs = _parse_components(item["outputs"])

                    functions[signature] = dict(
                        signature=signature,
                        name=item["name"],
                        inputs=inputs,
                        outputs=outputs,
                    )

                # parse contract events
                elif item["type"] == "event":
                    canonical, parameters = _parse_components(item["inputs"])
                    canonical = item["name"] + canonical
                    event_hash = Web3.sha3(text=canonical).hex()
                    signature = event_hash

                    events[signature] = dict(
                        signature=signature,
                        name=item["name"],
                        anonymous=item["anonymous"],
                        parameters=parameters,
                    )

        return dict(functions=functions, events=events)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from functools import lru_cache

from ..models.semantics_model import ContractSemantics
from ..semantics.router import Router

log = logging.getLogger(__name__)


@lru_cache(maxsize=128)
def amend_contract_semantics(semantics: ContractSemantics, router_=Router()):
    if semantics.code_hash in router_:
        try:
            semantics_updates = router_[semantics.code_hash]
            if "name" in semantics_updates:
                semantics.name = semantics_updates["name"]
            if "events" in semantics_updates:
                for signature, event_semantics in semantics_updates["events"].items():
                    semantics.events[signature] = event_semantics
            if "functions" in semantics_updates:
                for signature, function_semantics in semantics_updates[
                    "functions"
                ].items():
                    semantics.functions[signature] = function_semantics
            if "transformations" in semantics_updates:
                for signature, transformation in semantics_updates[
                    "transformations"
                ].items():
                    semantics.transformations[signature] = transformation

        except Exception as exc:
            log.warning(
                "Semantics update load for %s failed.",
                semantics.code_hash,
                exc_info=exc,
            )



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import (
    EventSemantics,
    FunctionSemantics,
    ParameterSemantics,
    TransformationSemantics,
)

erc721_transfer_event = EventSemantics(
    "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef",
    False,
    "Transfer",
    [
        ParameterSemantics("from", "address", [], True),
        ParameterSemantics("to", "address", [], True),
        ParameterSemantics("tokenId", "uint256", [], True),
    ],
)

erc721_transfer_event_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_approval_event = EventSemantics(
    "0x8c5be1e5ebec7d5bd14f71427d1e84f3dd0314c0f7b2291e5b200ac8c7c3b925",
    False,
    "Approval",
    [
        ParameterSemantics("owner", "address", [], True),
        ParameterSemantics("approved", "address", [], True),
        ParameterSemantics("tokenId", "uint256", [], True),
    ],
)

erc721_approval_event_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_approvalForAll_event = EventSemantics(
    "0x17307eab39ab6107e8899845ad3d59bd9653f200f220920489ca2b5937696c31",
    False,
    "ApprovalForAll",
    [
        ParameterSemantics("owner", "address", [], True),
        ParameterSemantics("operator", "address", [], True),
        ParameterSemantics("approved", "bool", [], False),
    ],
)

erc721_balanceOf_function = FunctionSemantics(
    "0x70a08231",
    "balanceOf",
    [ParameterSemantics("owner", "address", [])],
    [ParameterSemantics("", "uint256", [])],
)

erc721_ownerOf_function = FunctionSemantics(
    "0x6352211e",
    "ownerOf",
    [ParameterSemantics("tokenId", "uint256", [])],
    [ParameterSemantics("", "address", [])],
)

erc721_ownerOf_function_transformation = {
    "__input0__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input0__)"
    )
}

erc721_transferFrom_function = FunctionSemantics(
    "0x23b872dd",
    "transferFrom",
    [
        ParameterSemantics("from", "address", []),
        ParameterSemantics("to", "address", []),
        ParameterSemantics("tokenId", "uint256", []),
    ],
    [],
)

erc721_transferFrom_function_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_safeTransferFrom_function = FunctionSemantics(
    "0x42842e0e",
    "safeTransferFrom",
    [
        ParameterSemantics("from", "address", []),
        ParameterSemantics("to", "address", []),
        ParameterSemantics("tokenId", "uint256", []),
    ],
    [],
)

erc721_safeTransferFrom_function_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_safeTransferFrom_with_data_function = FunctionSemantics(
    "0xb88d4fde",
    "safeTransferFrom",
    [
        ParameterSemantics("from", "address", []),
        ParameterSemantics("to", "address", []),
        ParameterSemantics("tokenId", "uint256", []),
        ParameterSemantics("data", "bytes", [], dynamic=True),
    ],
    [],
)

erc721_safeTransferFrom_with_data_function_transformation = {
    "__input2__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input2__)"
    )
}

erc721_approve_function = FunctionSemantics(
    "0x095ea7b3",
    "approve",
    [
        ParameterSemantics("operator", "address", []),
        ParameterSemantics("tokenId", "uint256", []),
    ],
    [],
)

erc721_approve_function_transformation = {
    "__input1__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input1__)"
    )
}

erc721_setApprovalForAll_function = FunctionSemantics(
    "0xa22cb465",
    "setApprovalForAll",
    [
        ParameterSemantics("address", "address", []),
        ParameterSemantics("approved", "bool", []),
    ],
    [],
)

erc721_getApproved_function = FunctionSemantics(
    "0x081812fc",
    "getApproved",
    [ParameterSemantics("tokenId", "uint256", [])],
    [ParameterSemantics("", "address", [])],
)

erc721_getApproved_function_transformation = {
    "__input0__": TransformationSemantics(
        transformed_type="nft", transformation="decode_nft(__input0__)"
    )
}

erc721_isApprovedForAll_function = FunctionSemantics(
    "0xe985e9c5",
    "isApprovedForAll",
    [
        ParameterSemantics("owner", "address", []),
        ParameterSemantics("operator", "address", []),
    ],
    [ParameterSemantics("", "bool", [])],
)

ERC721_EVENTS = {
    erc721_transfer_event.signature: erc721_transfer_event,
    erc721_approval_event.signature: erc721_approval_event,
    erc721_approvalForAll_event.signature: erc721_approvalForAll_event,
}

ERC721_FUNCTIONS = {
    erc721_balanceOf_function.signature: erc721_balanceOf_function,
    erc721_ownerOf_function.signature: erc721_ownerOf_function,
    erc721_transferFrom_function.signature: erc721_transferFrom_function,
    erc721_safeTransferFrom_function.signature: erc721_safeTransferFrom_function,
    erc721_safeTransferFrom_with_data_function.signature: erc721_safeTransferFrom_with_data_function,
    erc721_approve_function.signature: erc721_approve_function,
    erc721_setApprovalForAll_function.signature: erc721_setApprovalForAll_function,
    erc721_getApproved_function.signature: erc721_getApproved_function,
    erc721_isApprovedForAll_function.signature: erc721_isApprovedForAll_function,
}

ERC721_TRANSFORMATIONS = {
    erc721_transfer_event.signature: erc721_transfer_event_transformation,
    erc721_approval_event.signature: erc721_approval_event_transformation,
    erc721_ownerOf_function.signature: erc721_ownerOf_function_transformation,
    erc721_transferFrom_function.signature: erc721_transferFrom_function_transformation,
    erc721_safeTransferFrom_function.signature: erc721_safeTransferFrom_function_transformation,
    erc721_safeTransferFrom_with_data_function.signature: erc721_safeTransferFrom_with_data_function_transformation,
    erc721_approve_function.signature: erc721_approve_function_transformation,
    erc721_getApproved_function.signature: erc721_getApproved_function_transformation,
}



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional


@dataclass
class BlockMetadata:
    block_number: int
    block_hash: str
    timestamp: datetime
    parent_hash: str
    miner: str
    gas_limit: int
    gas_used: int
    tx_count: int

    # for future use
    canonical: bool = True

    @staticmethod
    def from_raw(w3block) -> BlockMetadata:
        return w3block.to_object()


@dataclass
class TransactionMetadata:
    tx_hash: str
    block_number: int
    gas_price: int
    from_address: str
    to_address: str
    tx_index: int
    tx_value: int
    gas_limit: int
    gas_used: int
    success: bool

    # for future use
    gas_refund: int = None
    return_value: str = None
    exception_error: str = None
    exception_error_type: str = None
    revert_reason: str = None

    @staticmethod
    def from_raw(w3transaction, w3receipt) -> TransactionMetadata:
        return w3transaction.to_object(w3receipt)


@dataclass
class Event:
    contract: str
    topics: List[str]
    log_data: Optional[str]
    log_index: int

    # for future use
    call_id: str = None

    @staticmethod
    def from_raw(w3log) -> Event:
        return w3log.to_object()


@dataclass
class Call:
    call_type: str
    call_gas: int
    from_address: str
    to_address: str
    call_value: int
    call_data: str
    return_value: str
    gas_used: int
    status: bool
    error: str
    subcalls: Optional[List[Call]] = field(default_factory=list)

    # for future use
    call_id: str = None
    created_address: str = None
    gas_refund: int = None
    exception_error: str = None
    exception_error_type: str = None
    revert_reason: str = None
    success: bool = None

    @staticmethod
    def from_raw(w3calltree) -> Call:
        return w3calltree.to_object()


@dataclass
class Block:
    chain_id: str
    metadata: BlockMetadata
    transactions: List[Transaction]

    @staticmethod
    def from_raw(chain_id, w3block, w3transactions=None) -> Block:
        data = w3block.to_object()
        if w3transactions:
            transactions = [
                Transaction.from_raw(w3transaction, w3receipt, w3calltree)
                for (w3transaction, w3receipt, w3calltree) in w3transactions
            ]
        else:
            transactions = []

        return Block(chain_id=chain_id, metadata=data, transactions=transactions)


@dataclass
class Transaction:
    metadata: TransactionMetadata
    root_call: Call
    events: List[Event]

    @staticmethod
    def from_raw(w3transaction, w3receipt, w3calltree) -> Transaction:
        data = w3transaction.to_object(w3receipt)
        events = [w3log.to_object() for w3log in w3receipt.logs]
        root_call = w3calltree.to_object()
        return Transaction(metadata=data, root_call=root_call, events=events)



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import glob
import importlib
import os

from ..semantics.base import BaseType, Base


class Router:
    """
    Semantics router.
    Returns all objects withs semantics to include.
    """

    root_dir = os.path.dirname(__file__)
    root_module_name = ".".join(__name__.split(".")[:-1])

    def __new__(cls) -> BaseType:
        return cls._get_semantics()

    @classmethod
    def _get_semantics(cls) -> BaseType:
        """
        Get all available semantics.
        Match pattern:
            - .py file
            - object is a class type
            - object is a Base subclass
        """
        rv = {}
        files = (
            semantic
            for semantic in glob.iglob(cls.root_dir + "**/**", recursive=True)
            if os.path.isfile(semantic)
            and "__" not in semantic
            and semantic.endswith(".py")
        )

        for filename in files:
            filename = filename.replace("/", ".").replace(".py", "")
            imported_module = importlib.import_module(
                f"{cls.root_module_name}{filename.split(cls.root_module_name)[-1]}"
            )
            for item in dir(imported_module):
                obj = getattr(imported_module, item)
                if isinstance(obj, type) and issubclass(obj, Base) and obj != Base:
                    rv[obj.code_hash] = obj.contract_semantics

        return rv



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional, Any, List, Dict

from ethtx.models.decoded_model import DecodedCall, DecodedTransfer, Proxy
from ethtx.models.objects_model import (
    Block,
    Transaction,
    Call,
    Event,
    TransactionMetadata,
    BlockMetadata,
)
from ethtx.providers.semantic_providers import SemanticsRepository


class ABIBasic:
    def __init__(self, repository: SemanticsRepository, chain_id: str):
        self._default_chain = chain_id
        self._repository: SemanticsRepository = repository


class ABISubmoduleAbc(ABC, ABIBasic):
    @abstractmethod
    def decode(self, *args, **kwargs) -> Any:
        ...


class IABIDecoder(ABC, ABIBasic):
    def __init__(
        self,
        repository: SemanticsRepository,
        chain_id: str,
        strict: Optional[bool] = False,
    ):
        super().__init__(repository, chain_id)
        self.strict = strict

    @abstractmethod
    def decode_transaction(
        self, block: Block, transaction: Transaction, proxies: Dict[str, Proxy]
    ):
        ...

    @abstractmethod
    def decode_calls(
        self,
        call: Call,
        block: BlockMetadata,
        transaction: TransactionMetadata,
        proxies: Dict[str, Proxy],
    ) -> ABISubmoduleAbc.decode:
        ...

    @abstractmethod
    def decode_events(
        self,
        events: [Event],
        block: BlockMetadata,
        transaction: TransactionMetadata,
        proxies: Dict[str, Proxy],
    ) -> ABISubmoduleAbc.decode:
        ...

    @abstractmethod
    def decode_transfers(
        self, call: DecodedCall, events: [Event], proxies: Dict[str, Proxy]
    ) -> ABISubmoduleAbc.decode:
        ...

    @abstractmethod
    def decode_balances(
        self, transfers: List[DecodedTransfer]
    ) -> ABISubmoduleAbc.decode:
        ...



#  Copyright 2021 DAI Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ethtx.models.semantics_model import TransformationSemantics
from ethtx.semantics.base import Base


class DaiJoin(Base):
    code_hash = "0x76dac2e14412f6ee4c5ca566296c954598168738955fa974cb3302cb73b95f0f"

    contract_semantics = dict(
        transformations={
            "0x3b4da69f": {  # join
                "wad": TransformationSemantics(transformation="wad / 10**18")
            },
            "0xef693bed": {  # exit
                "wad": TransformationSemantics(transformation="wad / 10**18")
            },
        }
    )
